@@ -1,8 +1,7 @@
 from __future__ import annotations
 
-import asyncio
 from abc import ABC, abstractmethod
-from collections import OrderedDict
+from collections import OrderedDict, deque
 from collections.abc import (
     AsyncIterator,
     Awaitable,
@@ -12,17 +11,24 @@ from collections.abc import (
     Iterator,
     Mapping,
 )
-from contextlib import contextmanager, suppress
+from contextlib import asynccontextmanager, contextmanager, suppress
 from dataclasses import dataclass, field
 from enum import StrEnum
 from functools import partialmethod, singledispatchmethod, update_wrapper
-from inspect import Signature, isclass, iscoroutinefunction, markcoroutinefunction
+from inspect import (
+    Signature,
+    isasyncgenfunction,
+    isclass,
+    iscoroutinefunction,
+    isgeneratorfunction,
+    markcoroutinefunction,
+)
 from inspect import signature as inspect_signature
 from logging import Logger, getLogger
-from queue import Empty, Queue
 from types import MethodType
 from typing import (
     Any,
+    AsyncContextManager,
     ClassVar,
     ContextManager,
     Literal,
@@ -32,7 +38,6 @@ from typing import (
     overload,
     runtime_checkable,
 )
-from uuid import uuid4
 
 from injection._core.common.asynchronous import (
     AsyncCaller,
@@ -42,14 +47,22 @@ from injection._core.common.asynchronous import (
 )
 from injection._core.common.event import Event, EventChannel, EventListener
 from injection._core.common.invertible import Invertible, SimpleInvertible
-from injection._core.common.lazy import Lazy, LazyMapping
-from injection._core.common.threading import synchronized
-from injection._core.common.type import InputType, TypeInfo, get_return_types
+from injection._core.common.key import new_short_key
+from injection._core.common.lazy import Lazy, alazy, lazy
+from injection._core.common.type import (
+    InputType,
+    TypeInfo,
+    get_return_types,
+    get_yield_hint,
+)
 from injection._core.hook import Hook, apply_hooks
 from injection._core.injectables import (
+    AsyncCMScopedInjectable,
+    CMScopedInjectable,
     Injectable,
     ShouldBeInjectable,
     SimpleInjectable,
+    SimpleScopedInjectable,
     SingletonInjectable,
 )
 from injection.exceptions import (
@@ -198,14 +211,12 @@ class Record[T](NamedTuple):
 
 @dataclass(repr=False, eq=False, kw_only=True, slots=True)
 class Updater[T]:
-    factory: Caller[..., T]
     classes: Iterable[InputType[T]]
-    injectable_factory: InjectableFactory[T]
+    injectable: Injectable[T]
     mode: Mode
 
     def make_record(self) -> Record[T]:
-        injectable = self.injectable_factory(self.factory)
-        return Record(injectable, self.mode)
+        return Record(self.injectable, self.mode)
 
 
 @dataclass(repr=False, eq=False, frozen=True, slots=True)
@@ -262,7 +273,6 @@ class Locator(Broker):
     def __injectables(self) -> frozenset[Injectable[Any]]:
         return frozenset(record.injectable for record in self.__records.values())
 
-    @synchronized()
     def update[T](self, updater: Updater[T]) -> Self:
         updater = self.__update_preprocessing(updater)
         record = updater.make_record()
@@ -276,7 +286,6 @@ class Locator(Broker):
 
         return self
 
-    @synchronized()
     def unlock(self) -> Self:
         for injectable in self.__injectables:
             injectable.unlock()
@@ -285,6 +294,9 @@ class Locator(Broker):
 
     async def all_ready(self) -> None:
         for injectable in self.__injectables:
+            if injectable.is_locked:
+                continue
+
             await injectable.aget_instance()
 
     def add_listener(self, listener: EventListener) -> Self:
@@ -350,7 +362,7 @@ type PriorityStr = Literal["low", "high"]
 
 @dataclass(eq=False, frozen=True, slots=True)
 class Module(Broker, EventListener):
-    name: str = field(default_factory=lambda: f"anonymous@{uuid4().hex[:7]}")
+    name: str = field(default_factory=lambda: f"anonymous@{new_short_key()}")
     __channel: EventChannel = field(
         default_factory=EventChannel,
         init=False,
@@ -393,7 +405,7 @@ class Module(Broker, EventListener):
 
     @property
     def __brokers(self) -> Iterator[Broker]:
-        yield from tuple(self.__modules)
+        yield from self.__modules
         yield self.__locator
 
     def injectable[**P, T](
@@ -402,6 +414,7 @@ class Module(Broker, EventListener):
         /,
         *,
         cls: InjectableFactory[T] = SimpleInjectable,
+        ignore_type_hint: bool = False,
         inject: bool = True,
         on: TypeInfo[T] = (),
         mode: Mode | ModeStr = Mode.get_default(),
@@ -409,12 +422,11 @@ class Module(Broker, EventListener):
         def decorator(
             wp: Callable[P, T] | Callable[P, Awaitable[T]],
         ) -> Callable[P, T] | Callable[P, Awaitable[T]]:
-            factory = _get_caller(self.make_injected_function(wp) if inject else wp)
-            classes = get_return_types(wp, on)
+            factory = extract_caller(self.make_injected_function(wp) if inject else wp)
+            hints = on if ignore_type_hint else (wp, on)
             updater = Updater(
-                factory=factory,
-                classes=classes,
-                injectable_factory=cls,
+                classes=get_return_types(hints),
+                injectable=cls(factory),  # type: ignore[arg-type]
                 mode=Mode(mode),
             )
             self.update(updater)
@@ -424,12 +436,66 @@ class Module(Broker, EventListener):
 
     singleton = partialmethod(injectable, cls=SingletonInjectable)
 
+    def scoped[**P, T](
+        self,
+        scope_name: str,
+        /,
+        *,
+        inject: bool = True,
+        on: TypeInfo[T] = (),
+        mode: Mode | ModeStr = Mode.get_default(),
+    ) -> Any:
+        def decorator(
+            wrapped: Callable[P, T]
+            | Callable[P, Awaitable[T]]
+            | Callable[P, Iterator[T]]
+            | Callable[P, AsyncIterator[T]],
+        ) -> (
+            Callable[P, T]
+            | Callable[P, Awaitable[T]]
+            | Callable[P, Iterator[T]]
+            | Callable[P, AsyncIterator[T]]
+        ):
+            injectable_class: Callable[[Caller[P, Any], str], Injectable[T]]
+            wrapper: (
+                Callable[P, T]
+                | Callable[P, Awaitable[T]]
+                | Callable[P, ContextManager[T]]
+                | Callable[P, AsyncContextManager[T]]
+            )
+
+            if isasyncgenfunction(wrapped):
+                hint = get_yield_hint(wrapped)
+                injectable_class = AsyncCMScopedInjectable
+                wrapper = asynccontextmanager(wrapped)
+
+            elif isgeneratorfunction(wrapped):
+                hint = get_yield_hint(wrapped)
+                injectable_class = CMScopedInjectable
+                wrapper = contextmanager(wrapped)
+
+            else:
+                injectable_class = SimpleScopedInjectable
+                hint = wrapper = wrapped  # type: ignore[assignment]
+
+            hints = on if hint is None else (hint, on)
+            self.injectable(
+                wrapper,
+                cls=lambda factory: injectable_class(factory, scope_name),
+                ignore_type_hint=True,
+                inject=inject,
+                on=hints,
+                mode=mode,
+            )
+            return wrapped
+
+        return decorator
+
     def should_be_injectable[T](self, wrapped: type[T] | None = None, /) -> Any:
         def decorator(wp: type[T]) -> type[T]:
             updater = Updater(
-                factory=SyncCaller(wp),
                 classes=(wp,),
-                injectable_factory=lambda _: ShouldBeInjectable(wp),
+                injectable=ShouldBeInjectable(wp),
                 mode=Mode.FALLBACK,
             )
             self.update(updater)
@@ -446,9 +512,10 @@ class Module(Broker, EventListener):
         mode: Mode | ModeStr = Mode.get_default(),
     ) -> Any:
         def decorator(wp: type[T]) -> type[T]:
-            lazy_instance = Lazy(wp)
+            lazy_instance = lazy(wp)
             self.injectable(
                 lambda: ~lazy_instance,
+                ignore_type_hint=True,
                 inject=False,
                 on=(wp, on),
                 mode=mode,
@@ -465,14 +532,12 @@ class Module(Broker, EventListener):
         alias: bool = False,
         mode: Mode | ModeStr = Mode.get_default(),
     ) -> Self:
-        if not alias:
-            cls = type(instance)
-            on = (cls, on)
-
+        hints = on if alias else (type(instance), on)
         self.injectable(
             lambda: instance,
+            ignore_type_hint=True,
             inject=False,
-            on=on,
+            on=hints,
             mode=mode,
         )
         return self
@@ -504,7 +569,7 @@ class Module(Broker, EventListener):
     def make_injected_function(self, wrapped, /):  # type: ignore[no-untyped-def]
         metadata = InjectMetadata(wrapped)
 
-        @metadata.on_setup
+        @metadata.task
         def listen() -> None:
             metadata.update(self)
             self.add_listener(metadata)
@@ -582,12 +647,10 @@ class Module(Broker, EventListener):
 
     def aget_lazy_instance(self, cls, default=None, *, cache=False):  # type: ignore[no-untyped-def]
         if cache:
-            coroutine = self.aget_instance(cls, default)
-            return asyncio.ensure_future(coroutine)
+            return alazy(lambda: self.aget_instance(cls, default))
 
         function = self.make_injected_function(lambda instance=default: instance)
-        metadata = function.__inject_metadata__
-        metadata.set_owner(cls)
+        metadata = function.__inject_metadata__.set_owner(cls)
         return SimpleAwaitable(metadata.acall)
 
     @overload
@@ -610,11 +673,10 @@ class Module(Broker, EventListener):
 
     def get_lazy_instance(self, cls, default=None, *, cache=False):  # type: ignore[no-untyped-def]
         if cache:
-            return Lazy(lambda: self.get_instance(cls, default))
+            return lazy(lambda: self.get_instance(cls, default))
 
         function = self.make_injected_function(lambda instance=default: instance)
-        metadata = function.__inject_metadata__
-        metadata.set_owner(cls)
+        metadata = function.__inject_metadata__.set_owner(cls)
         return SimpleInvertible(metadata.call)
 
     def update[T](self, updater: Updater[T]) -> Self:
@@ -670,8 +732,11 @@ class Module(Broker, EventListener):
         priority: Priority | PriorityStr = Priority.get_default(),
     ) -> Iterator[None]:
         self.use(module, priority=priority)
-        yield
-        self.stop_using(module)
+
+        try:
+            yield
+        finally:
+            self.stop_using(module)
 
     def change_priority(self, module: Module, priority: Priority | PriorityStr) -> Self:
         priority = Priority(priority)
@@ -682,7 +747,6 @@ class Module(Broker, EventListener):
 
         return self
 
-    @synchronized()
     def unlock(self) -> Self:
         for broker in self.__brokers:
             broker.unlock()
@@ -714,12 +778,14 @@ class Module(Broker, EventListener):
         self.__check_locking()
 
         with self.__channel.dispatch(event):
-            yield
-            message = str(event)
-            self.__debug(message)
+            try:
+                yield
+            finally:
+                message = str(event)
+                self.__debug(message)
 
     def __debug(self, message: object) -> None:
-        for logger in tuple(self.__loggers):
+        for logger in self.__loggers:
             logger.debug(message)
 
     def __check_locking(self) -> None:
@@ -741,10 +807,8 @@ class Module(Broker, EventListener):
         with suppress(KeyError):
             return cls.__instances[name]
 
-        with synchronized():
-            instance = cls(name)
-            cls.__instances[name] = instance
-
+        instance = cls(name)
+        cls.__instances[name] = instance
         return instance
 
     @classmethod
@@ -766,10 +830,7 @@ InjectedFunction
 
 @dataclass(repr=False, frozen=True, slots=True)
 class Dependencies:
-    mapping: Mapping[str, Injectable[Any]]
-
-    def __bool__(self) -> bool:
-        return bool(self.mapping)
+    lazy_mapping: Lazy[Mapping[str, Injectable[Any]]]
 
     def __iter__(self) -> Iterator[tuple[str, Any]]:
         for name, injectable in self.mapping.items():
@@ -783,10 +844,11 @@ class Dependencies:
 
     @property
     def are_resolved(self) -> bool:
-        if isinstance(self.mapping, LazyMapping) and not self.mapping.is_set:
-            return False
+        return self.lazy_mapping.is_set
 
-        return bool(self)
+    @property
+    def mapping(self) -> Mapping[str, Injectable[Any]]:
+        return ~self.lazy_mapping
 
     async def aget_arguments(self) -> dict[str, Any]:
         return {key: value async for key, value in self}
@@ -795,12 +857,13 @@ class Dependencies:
         return dict(self)
 
     @classmethod
-    def from_mapping(cls, mapping: Mapping[str, Injectable[Any]]) -> Self:
-        return cls(mapping)
+    def from_iterable(cls, iterable: Iterable[tuple[str, Injectable[Any]]]) -> Self:
+        lazy_mapping = Lazy(lambda: dict(iterable))
+        return cls(lazy_mapping)
 
     @classmethod
     def empty(cls) -> Self:
-        return cls.from_mapping({})
+        return cls.from_iterable(())
 
     @classmethod
     def resolve(
@@ -809,8 +872,8 @@ class Dependencies:
         module: Module,
         owner: type | None = None,
     ) -> Self:
-        dependencies = LazyMapping(cls.__resolver(signature, module, owner))
-        return cls.from_mapping(dependencies)
+        iterable = cls.__resolver(signature, module, owner)
+        return cls.from_iterable(iterable)
 
     @classmethod
     def __resolver(
@@ -851,21 +914,21 @@ class InjectMetadata[**P, T](Caller[P, T], EventListener):
     __slots__ = (
         "__dependencies",
         "__owner",
-        "__setup_queue",
         "__signature",
+        "__tasks",
         "__wrapped",
     )
 
     __dependencies: Dependencies
     __owner: type | None
-    __setup_queue: Queue[Callable[..., Any]] | None
     __signature: Signature
+    __tasks: deque[Callable[..., Any]]
     __wrapped: Callable[P, T]
 
     def __init__(self, wrapped: Callable[P, T], /) -> None:
         self.__dependencies = Dependencies.empty()
         self.__owner = None
-        self.__setup_queue = Queue()
+        self.__tasks = deque()
         self.__wrapped = wrapped
 
     @property
@@ -873,10 +936,8 @@ class InjectMetadata[**P, T](Caller[P, T], EventListener):
         with suppress(AttributeError):
             return self.__signature
 
-        with synchronized():
-            signature = inspect_signature(self.wrapped, eval_str=True)
-            self.__signature = signature
-
+        signature = inspect_signature(self.wrapped, eval_str=True)
+        self.__signature = signature
         return signature
 
     @property
@@ -900,12 +961,12 @@ class InjectMetadata[**P, T](Caller[P, T], EventListener):
         return self.__bind(args, kwargs, additional_arguments)
 
     async def acall(self, /, *args: P.args, **kwargs: P.kwargs) -> T:
-        self.__setup()
+        self.__run_tasks()
         arguments = await self.abind(args, kwargs)
         return self.wrapped(*arguments.args, **arguments.kwargs)
 
     def call(self, /, *args: P.args, **kwargs: P.kwargs) -> T:
-        self.__setup()
+        self.__run_tasks()
         arguments = self.bind(args, kwargs)
         return self.wrapped(*arguments.args, **arguments.kwargs)
 
@@ -921,19 +982,13 @@ class InjectMetadata[**P, T](Caller[P, T], EventListener):
         self.__owner = owner
         return self
 
-    @synchronized()
     def update(self, module: Module) -> Self:
         self.__dependencies = Dependencies.resolve(self.signature, module, self.__owner)
         return self
 
-    def on_setup[**_P, _T](self, wrapped: Callable[_P, _T] | None = None, /) -> Any:
+    def task[**_P, _T](self, wrapped: Callable[_P, _T] | None = None, /) -> Any:
         def decorator(wp: Callable[_P, _T]) -> Callable[_P, _T]:
-            queue = self.__setup_queue
-
-            if queue is None:
-                raise RuntimeError(f"`{self}` is already up.")
-
-            queue.put_nowait(wp)
+            self.__tasks.append(wp)
             return wp
 
         return decorator(wrapped) if wrapped else decorator
@@ -964,26 +1019,10 @@ class InjectMetadata[**P, T](Caller[P, T], EventListener):
         bound.arguments = bound.arguments | additional_arguments | bound.arguments
         return Arguments(bound.args, bound.kwargs)
 
-    def __close_setup_queue(self) -> None:
-        self.__setup_queue = None
-
-    def __setup(self) -> None:
-        queue = self.__setup_queue
-
-        if queue is None:
-            return
-
-        while True:
-            try:
-                task = queue.get_nowait()
-            except Empty:
-                break
-
+    def __run_tasks(self) -> None:
+        while tasks := self.__tasks:
+            task = tasks.popleft()
             task()
-            queue.task_done()
-
-        queue.join()
-        self.__close_setup_queue()
 
 
 class InjectedFunction[**P, T](ABC):
@@ -1037,11 +1076,13 @@ class SyncInjectedFunction[**P, T](InjectedFunction[P, T]):
         return self.__inject_metadata__.call(*args, **kwargs)
 
 
-def _get_caller[**P, T](function: Callable[P, T]) -> Caller[P, T]:
+def extract_caller[**P, T](
+    function: Callable[P, T] | Callable[P, Awaitable[T]],
+) -> Caller[P, T]:
     if iscoroutinefunction(function):
         return AsyncCaller(function)
 
     elif isinstance(function, InjectedFunction):
         return function.__inject_metadata__
 
-    return SyncCaller(function)
+    return SyncCaller(function)  # type: ignore[arg-type]
